{$V-,A+,S+,R+,D+}
Unit Com;
  { Unit with high level async I/O routines using the fossil driver }

INTERFACE

Uses DOS;

{$I KEYCODES.INC} {ASCII & extended key codes + color & mode codes}
Type
  ExtendedKeyProc = Procedure(ScanCode: byte);
  IdleKeyProc = Procedure;

Const
  Local: boolean = true;
  Quiet: boolean = true;
  UseColors: boolean = true;

  CheckDCD: boolean = true;
  ComError: byte = 0; { 1=carrier lost, 2=timed out }

  KeyIdleTimeout: word = 300;
  KeyIdleWarning: word = 120;

  MorePrompt: string[15] = 'More (Y/n)? ';

Var
  ExtendedKeyHook: ExtendedKeyProc;
  IdleKeyHook: IdleKeyProc;
  ExtendedKey: boolean;
  LocalKey: boolean;
  Escaped: boolean;

Function InitComPort(Port: byte; Baud: word; CharLength, Parity,
                     StopBits: byte; SoftFlow, HardFlow: boolean): boolean;
Procedure SetExtendedKeyHook(EKP: ExtendedKeyProc);
Procedure SetIdleKeyHook(IKP: IdleKeyProc);

Procedure ComChar(C: Char);
Procedure ComWrite(S: string);
Procedure ComWriteLn(S: string);
Procedure ComLn;

Procedure ComRead(var gstr: string; max: byte; reset: boolean);
Procedure ComReadLn(var gstr: string; max: byte; reset: boolean);
Function ComReadKey : char;

Procedure TypeFile(FileName: pathstr; PauseLines: byte);
Procedure Chat(LocalColor, RemoteColor: byte);
Procedure ATCommand(S: string);

Function ANSITextColor(color:byte): string;
Function ANSITextBackground(color:byte): string;
Procedure TextColor(color: byte);
Procedure TextBackground(color: byte);
Procedure GotoXY(col, row: byte);
Procedure WhereXY(var col, row: byte);
Procedure SaveXY;
Procedure RestoreXY;
Procedure GoUp(rows: byte);
Procedure GoDown(rows: byte);
Procedure GoRight(cols: byte);
Procedure GoLeft(cols: byte);
Procedure ClrEol;
Procedure ClrBol;
Procedure ClrScr;
Procedure TextMode(mode: byte);

IMPLEMENTATION

Uses GiveTime, TxtShare, Timer, Fossil, ANSI;

Const
  ExitSave: Pointer = NiL;
  AnsiColors : array[0..7] of char = ('0','4','2','6','1','5','3','7');

Var
  Regs: Registers;
  IT: EventTimer;

Function IntStr(i: longint): string;
var istr: string;
begin
  Str(i,istr);
  IntStr:= istr;
end;

{$F+}
Procedure DefExtKeyHook(SC: byte);
begin
end;
{$F-}

{$F+}
Procedure DefIdleKeyHook;
begin
end;
{$F-}

Procedure SetExtendedKeyHook(EKP: ExtendedKeyProc);
begin
  ExtendedKeyHook:= EKP;
end;

Procedure SetIdleKeyHook(IKP: IdleKeyProc);
begin
  IdleKeyHook:= IKP;
end;

Procedure ComExitProc; Far;
begin
  ExitProc:= ExitSave;
  FosClose;
end;

Function InitComPort(Port: byte; Baud: word; CharLength, Parity,
                     StopBits: byte; SoftFlow, HardFlow: boolean): boolean;
var linemask, flowmask: byte;
begin
  linemask:= $00;
  flowmask:= $F0;
  case port of
    1: fosport:= 0; {Com1}
    2: fosport:= 1; {Com2}
    3: fosport:= 2; {Com3}
    4: fosport:= 3; {Com4}
  end;
  case baud of
      300: linemask:= linemask or $40;
      600: linemask:= linemask or $60;
     1200: linemask:= linemask or $80;
     2400: linemask:= linemask or $A0;
     4800: linemask:= linemask or $C0;
     9600: linemask:= linemask or $E0;
    19200: linemask:= linemask or $00;
    38400: linemask:= linemask or $20;
  end;
  case Parity of
    0: linemask:= linemask or $00; {none}
    1: linemask:= linemask or $08; {odd}
    2: linemask:= linemask or $18; {even}
  end;
  case StopBits of
    1: linemask:= linemask or $00; {1 stop bit}
    2: linemask:= linemask or $04; {1.5 stop bits for CharLenght = 5 and 2 for others}
  end;
  case CharLength of
    5: linemask:= linemask or $00;
    6: linemask:= linemask or $01;
    7: linemask:= linemask or $02;
    8: linemask:= linemask or $03;
  end;
  SetBaudRate(linemask);

  if SoftFlow then
    flowmask:= flowmask or $09;
  if HardFlow then
    flowmask:= flowmask or $02;
  SetFosFlow(flowmask);

  if FosOpen then
    begin
      ExitSave:= ExitProc;
      ExitProc:= @ComExitProc;
      InitComPort:= true;
    end
  else
    InitComPort:= false;
end;

Procedure ComChar(C: Char);
var trans: boolean;
begin
  if not local then
    begin
      PutCharNow(C,trans);
      while not trans do
        begin
          GiveTimeSlice;
          PutCharNow(C,trans);
        end;
    end;
  if (not Quiet) or (C <> Bell) then
    ANSIChar(C);
end;

Procedure ComWrite(S: string);
var i: byte;
begin
  for i:= 1 to length(S) do
    ComChar(S[i]);
end;

Procedure ComWriteLn(S: string);
var i: byte;
begin
  for i:= 1 to length(S) do
    ComChar(S[i]);
  ComChar(CR);
  ComChar(LF);
end;

Procedure ComLn;
begin
  ComChar(CR);
  ComChar(LF);
end;

Function ComReadKey : char;
var
  rkey : char;
  lkey : word;
  Warned : boolean;

  Procedure InactivityBell;
  var b: byte;
  begin
    for b:= 1 to 3 do
      begin
        ComWrite(Bell);
        delay(300);
      end;
  end;

begin
  rkey:= null;
  NewTimerSecs(IT,KeyIdleTimeout-KeyIdleWarning);
  Warned:= false;
  ExtendedKey:= false;

  repeat
    if CheckDCD and (not Carrier) and (not local) then
      begin
        if not quiet then writeln(Bell);
        writeln('Carrier lost... Terminating!');
        ComError:= 1;
        Halt;
      end;
    if TimerExpired(IT) then
      if not Warned then
        begin
          InactivityBell;
          NewTimerSecs(IT,KeyIdleWarning);
          Warned := true;
        end
      else
        begin
          ComWriteLn(Bell+'Inactivity timeout... Terminating!');
          DropCarrier;
          ComError:= 2;
          Halt;
        end;

    IdleKeyHook;
    GiveTimeSlice;

    PeakKbd(lkey,LocalKey);
      if LocalKey then
        begin
          GetKbd(lkey);
          if Lo(lkey) = 0 then
            begin
              ExtendedKey:= True;
              ExtendedKeyHook(Hi(lkey));
            end
          else
            rkey := Chr(Lo(lkey));
        end
      else
        if not Local then
          if InputReady then
            GetChar(rkey);

  until rkey <> null;
  ComReadKey := rkey;
end;

Procedure ComRead(var gstr: string; max: byte; reset: boolean);
var
  gchar : char;
begin
  if reset then gstr:= '';
  if length(gstr) > max then
    gstr[0] := chr(max);
  if gstr <> '' then
    ComWrite(gstr);
  repeat
    gchar := ComReadKey;
    if (gchar in [#1..#7,#9..#12,#14..#26,#28..#255]) and (length(gstr) < max) then
      begin
        gstr := gstr + gchar;
        ComChar(gchar);
      end
    else if (gchar = #8) and (length(gstr) > 0) then
      begin
        delete(gstr,length(gstr),1);
	ComWrite(#8+#32+#8)
      end
  until gchar in [#13,#27];
  if gchar = #27 then
    begin
      gstr:= '';
      Escaped:= true;
    end
  else
    Escaped:= false;
end;

Procedure ComReadLn(var gstr: string; max: byte; reset: boolean);
var
  gchar : char;
begin
  if reset then gstr:= '';
  if length(gstr) > max then
    gstr[0] := chr(max);
  if gstr <> '' then
    ComWrite(gstr);
  repeat
    gchar := ComReadKey;
    if (gchar in [#1..#7,#9..#12,#14..#26,#28..#255]) and (length(gstr) < max) then
      begin
        gstr := gstr + gchar;
        ComChar(gchar);
      end
    else if (gchar = #8) and (length(gstr) > 0) then
      begin
        delete(gstr,length(gstr),1);
	ComWrite(#8+#32+#8)
      end
  until gchar in [#13,#27];
  if gchar = #27 then
    begin
      gstr:= '';
      Escaped:= true;
    end
  else
    Escaped:= false;
  ComLn;
end;

Procedure Chat(LocalColor, RemoteColor: byte);
var
  Quit: boolean;
  Ch: char;
  LastLocal: boolean;
begin
  Quit:= false;
  LastLocal:= not localkey;
  repeat
    Ch := ComReadKey;
    if UseColors then
      begin
        if localkey and (not lastlocal) then TextColor(LocalColor) else
          if (not localkey) and lastlocal then
            TextColor(RemoteColor);
      end;
    if localkey then lastlocal:= true else
      lastlocal:= false;
    case Ch of
      #13 : ComLn;
      #27 : if localkey then Quit := true;
      #8  : ComWrite(#8+#32+#8);
      #32..#255 : ComChar(Ch);
    end;
  until Quit;
end;

Procedure ATCommand(S: string);
var i: byte;
begin
  for i:= 1 to length(S) do
    begin
      case S[i] of
      '|': begin
             ComLn;
             Delay(500);
           end;
      '~': Delay(500);
      else
        begin
          ComChar(S[i]);
          Delay(10);
        end;
      end;
    end;
end;

Function ANSITextColor(color:byte): string;
var EscStr: string;
begin
  EscStr:= Esc+'[';
  if color > 128 then
    begin
      color:= color-128;
      EscStr:= EscStr+'5;'
    end
  else
    EscStr:= EscStr+'0;';
  if color > 7 then
    begin
      color:= color-8;
      EscStr:= EscStr+'1;';
    end;
  EscStr:= EscStr+'3'+AnsiColors[color]+'m';
  ANSITextColor:= EscStr;
end;

Function ANSITextBackground(color:byte): string;
begin
  ANSITextBackground:= Esc+'[4'+AnsiColors[color]+'m';
end;

Procedure TextColor(color: byte);
begin
  if UseColors then
    ComWrite(ANSITextColor(color));
end;

Procedure TextBackground(color: byte);
begin
  if UseColors then
    ComWrite(ANSITextBackground(color));
end;

Procedure GoUp(rows: byte);
begin
  ComWrite(Esc+'['+intstr(rows)+'A');
end;

Procedure GoDown(rows: byte);
begin
  ComWrite(Esc+'['+intstr(rows)+'B');
end;

Procedure GoRight(cols: byte);
begin
  ComWrite(Esc+'['+intstr(cols)+'C');
end;

Procedure GoLeft(cols: byte);
begin
  ComWrite(Esc+'['+intstr(cols)+'D');
end;

Procedure GotoXY(col, row: byte);
begin
  ComWrite(Esc+'['+intstr(row)+';'intstr(col)+'H');
end;

Procedure WhereXY(var col, row: byte);
var x, y: byte;
begin
  GetFosCursor(y, x);
  inc(x);
  inc(y);
  col:= x;
  row:= y;
end;

Procedure SaveXY;
begin
  ComWrite(Esc+'[s');
end;

Procedure RestoreXY;
begin
  ComWrite(Esc+'[u');
end;

Procedure ClrEol;
begin
  ComWrite(Esc+'[0J');
end;

Procedure ClrBol;
begin
  ComWrite(Esc+'[1J');
end;

Procedure ClrScr;
begin
  ComWrite(Esc+'[2J');
end;

Procedure TextMode(mode: byte);
begin
  ComWrite(Esc+'[='+intstr(mode)+'h');
end;

Procedure TypeFile(FileName: pathstr; PauseLines: byte);
  {-No pause prompt, just set PauseLines = 0}
  var
    i: byte;
    OutFile: Text;

  Function Exist(FileName: pathstr): boolean;
  var
    Inf: SearchRec;
  begin
    FindFirst(FileName,(AnyFile and not Directory),Inf);
    Exist := (DOSError = 0);
  end;

  Function ShareReset: boolean;
    var
      OpenAttempts : byte;
      GoAhead: boolean;
    begin
      OpenAttempts:= 0;
      repeat
        {$I-} reset(outfile) {$I+};
        GoAhead:= (ioresult = 0);
        inc(OpenAttempts);
      until (GoAhead) OR (OpenAttempts > 10);
      ShareReset:= GoAhead;
    end;

  Procedure SendOut(filename: pathstr);
    var
      C : char;
    begin
      ShareAssign(outfile,filename,true,true);
      if ShareReset then
        begin
          i := 0;
	  while not Eof(outfile) do
	    begin
              Read(outfile,C);
              ComChar(C);
	      if Eoln(outfile) then
	        begin
                  ReadLn(outfile);
                  ComLn;
		  inc(i);
		  if (PauseLines <> 0) and (i >= PauseLines) then
		    begin
                      if CheckDCD and (not Carrier) and (not Local) then
	                begin
		          close(outfile);
                          exit;
		        end;
                      savexy;
		      ComWrite(MorePrompt);
		      C := ComReadKey;
		      if upcase(C) = 'N' then
		      begin
		        ComLn;
		        close(outfile);
		        exit;
		      end
		    else
                      begin
                        restorexy;
			i := 0;
                      end;
		    end;
		end;
	    end;
	  close(outfile);
        end;
    end;

  begin {TypeFile}
    if exist(FileName) then
      SendOut(FileName)
    else
      if UseColors and exist(FileName+'.ANS') then
        SendOut(FileName+'.ANS')
      else
        if exist(FileName+'.ASC') then
          SendOut(FileName+'.ASC');
  end; {TypeFile}

begin {ComUnit}
  SetExtendedKeyHook(DefExtKeyHook);
  SetIdleKeyHook(DefIdleKeyHook);
end. {ComUnit}

